#import sys
#sys.path.append("/home/seolho/paper/meent")

import torch
import torch.nn as nn
from meent.on_torch.convolution_matrix import to_conv_mat
from meent.rcwa import call_solver

import numpy as np
import copy

fourier_order = 2
mode_key = 2
dtype = 0 
device = 0

'''
class Optimizer:
    def __init__(self, optimizer, mee, target_name, lr = 0.001):
        self.mee = mee
        self.target_name = target_name
        self.target = getattr(self.mee, target_name)
        
        if not isinstance(target, torch.Tensor):
            self.target = torch.Tensor(self.target)
        self.target.requires_grad = True
        
        if isinstance(optimizer,str):
            optimizer = getattr(torch.optim, optimizer)
        self.optimizer = optimizer([self.target], lr = lr)
        
    def optimize(self, iterations = 1000):
        for iteration in range(iterations):
            E_conv_all = to_conv_mat(self.mee.ucell, fourier_order)
            o_E_conv_all = to_conv_mat(1 / self.mee.ucell, fourier_order)

            de_ri, de_ti, _, _, _ = self.mee.solve(self.mee.wavelength, E_conv_all, o_E_conv_all)

            self.optimizer.zero_grad()
            loss = self.loss(de_ti)
            loss.backward()
            self.optimizer.step()
            
            setattr(self.mee, self.target_name, self.target)
            
            print(loss)
            
    def loss(self, de_ti):
        return -de_ti[3, 2]
'''

class Optimizer:
    def __init__(self, optimizer, solver, target, lr = 0.001):
        self.solver = solver
        
        self.target = target
        if not isinstance(target, torch.Tensor):
            self.target = torch.Tensor(self.target)
        self.target.requires_grad = True
        
        if isinstance(optimizer,str):
            optimizer = getattr(torch.optim, optimizer)
        self.optimizer = optimizer([self.target], lr = lr)
        
    def optimize(self, iterations = 1000):
        for iteration in range(iterations):
            E_conv_all = to_conv_mat(self.solver.ucell, fourier_order)
            o_E_conv_all = to_conv_mat(1 / self.solver.ucell, fourier_order)

            de_ri, de_ti, _, _, _ = self.solver.solve(self.solver.wavelength, E_conv_all, o_E_conv_all)

            self.optimizer.zero_grad()
            loss = self.loss(de_ti)
            loss.backward()
            self.optimizer.step()
            
            print(loss)
            
    def loss(self, de_ti):
        return -de_ti[3, 2]
    
def load_setting(mode_key, dtype, device):
    grating_type = 2

    pol = 1  # 0: TE, 1: TM

    n_I = 1  # n_incidence
    n_II = 1  # n_transmission

    theta = 0
    phi = 0
    psi = 0 if pol else 90

    wavelength = 900

    ucell_materials = [1, 3.48]
    fourier_order = 2

    thickness, period = [1120.], [1000, 1000]

    ucell = np.array(
        [[
            [3., 1., 1., 1., 3.],
            [3., 1., 1., 1., 3.],
            [3., 1., 1., 1., 3.],
            [3., 1., 1., 1., 3.],
            [3., 1., 1., 1., 3.],
        ]]
    )

    if mode_key == 0:
        device = 0
        type_complex = np.complex128 if dtype == 0 else np.complex64
        ucell = ucell.astype(type_complex)

    elif mode_key == 1:  # JAX
        jax.config.update('jax_platform_name', 'cpu') if device == 0 else jax.config.update('jax_platform_name', 'gpu')

        if dtype == 0:
            jax.config.update("jax_enable_x64", True)
            type_complex = jnp.complex128
            ucell = ucell.astype(jnp.float64)
            ucell = jnp.array(ucell, dtype=jnp.float64)

        else:
            type_complex = jnp.complex64
            ucell = ucell.astype(jnp.float32)
            ucell = jnp.array(ucell, dtype=jnp.float32)


    else:  # Torch
        device = torch.device('cpu') if device == 0 else torch.device('cuda')
        type_complex = torch.complex128 if dtype == 0 else torch.complex64

        if dtype == 0:
            ucell = torch.tensor(ucell, dtype=torch.float64, device=device)
        else:
            ucell = torch.tensor(ucell, dtype=torch.float32, device=device)

    return grating_type, pol, n_I, n_II, theta, phi, psi, wavelength, thickness, ucell_materials, period, fourier_order,\
           type_complex, device, ucell



grating_type, pol, n_I, n_II, theta, phi, psi, wavelength, thickness, ucell_materials, period, fourier_order, \
type_complex, device, ucell = load_setting(mode_key, dtype, device)

ucell.requires_grad = True

solver = call_solver(mode_key, grating_type=grating_type, pol=pol, n_I=n_I, n_II=n_II, theta=theta, phi=phi,
                     psi=psi, fourier_order=fourier_order, wavelength=wavelength, period=period, ucell=ucell,
                     ucell_materials=ucell_materials, thickness=thickness, device=device,
                     type_complex=type_complex, )


optim = Optimizer('Adam', solver, solver.ucell, lr = 0.001)
optim.optimize()


import random
from typing import Tuple

import optax
import jax.numpy as jnp
import jax
import numpy as np

BATCH_SIZE = 5
NUM_TRAIN_STEPS = 1_000
RAW_TRAINING_DATA = np.random.randint(255, size=(NUM_TRAIN_STEPS, BATCH_SIZE, 1))

TRAINING_DATA = np.unpackbits(RAW_TRAINING_DATA.astype(np.uint8), axis=-1)
LABELS = jax.nn.one_hot(RAW_TRAINING_DATA % 2, 2).astype(jnp.float32).reshape(NUM_TRAIN_STEPS, BATCH_SIZE, 2)

initial_params = {
    'hidden': jax.random.normal(shape=[8, 32], key=jax.random.PRNGKey(0)),
    'output': jax.random.normal(shape=[32, 2], key=jax.random.PRNGKey(1)),
}


def net_(x: jnp.ndarray, params: jnp.ndarray) -> jnp.ndarray:
    x = jnp.dot(x, params['hidden'])
    x = jax.nn.relu(x)
    x = jnp.dot(x, params['output'])
    return x


def net(x: jnp.ndarray, params: jnp.ndarray) -> jnp.ndarray:
    x = jnp.dot(x, params['hidden'])
    x = jax.nn.relu(x)
    x = jnp.dot(x, params['output'])
    return x


def loss(params: optax.Params, batch: jnp.ndarray, labels: jnp.ndarray) -> jnp.ndarray:
    y_hat = net(batch, params)

    # optax also provides a number of common loss functions.
    loss_value = optax.sigmoid_binary_cross_entropy(y_hat, labels).sum(axis=-1)

    return loss_value.mean()


def fit(params: optax.Params, optimizer: optax.GradientTransformation) -> optax.Params:
    opt_state = optimizer.init(params)

    @jax.jit
    def step(params, opt_state, batch, labels):
        loss_value, grads = jax.value_and_grad(loss)(params, batch, labels)
        updates, opt_state = optimizer.update(grads, opt_state, params)
        params = optax.apply_updates(params, updates)
        return params, opt_state, loss_value

    for i, (batch, labels) in enumerate(zip(TRAINING_DATA, LABELS)):
        params, opt_state, loss_value = step(params, opt_state, batch, labels)
        if i % 100 == 0:
            print(f'step {i}, loss: {loss_value}')

    return params


# Finally, we can fit our parametrized function using the Adam optimizer
# provided by optax.
optimizer = optax.adam(learning_rate=1e-2)
params = fit(initial_params, optimizer)


if __name__ == '__main__':
    import meent.testcase

    mode = 1
    dtype = 0
    device = 0
    grating_type = 2

    conditions = meent.testcase.load_setting(mode, dtype, device, grating_type)

    initial_params = {
        # 'hidden': jax.random.normal(shape=[8, 32], key=jax.random.PRNGKey(0)),
        # 'output': jax.random.normal(shape=[32, 2], key=jax.random.PRNGKey(1)),
        'ucell': conditions['ucell']
    }

    optimizer = optax.adam(learning_rate=1e-2)
    params = fit(initial_params, optimizer)

    print(1)
    #######################
    aa = OptimizerTorch(**conditions)
    import meent.on_torch.optimizer.loss

    pois = ['ucell', 'thickness']
    parameters_to_fit = [(getattr(aa, poi)) for poi in pois]
    forward = aa.conv_solve
    loss_fn = meent.on_torch.optimizer.loss.LossDeflector(x_order=0, y_order=1)

    grad = aa.grad(pois, forward, loss_fn)
    print(1, grad)

    # case 1
    opt = torch.optim.SGD(parameters_to_fit, lr=1E-2)
    aa.fit(pois, forward, loss_fn, opt)
    print(3, grad)

    # case 2
    opt_algo = 'sgd'
    opt_kwargs = {'lr': 1E-2}
    aa.fit_general(pois, forward, loss_fn, opt_algo, opt_kwargs)
    print(3, grad)



import time
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np

from ._base import _BaseRCWA
from .convolution_matrix import to_conv_mat_discrete, put_permittivity_in_ucell, read_material_table, \
    to_conv_mat_continuous
from .field_distribution import field_dist_1d, field_dist_1d_conical, field_dist_2d, field_plot


class RCWAJax(_BaseRCWA):
    def __init__(self,
                 n_I=1.,
                 n_II=1.,
                 theta=0,
                 phi=0,
                 psi=0,
                 period=(100, 100),
                 wavelength=900,
                 ucell=None,
                 thickness=None,
                 mode=1,
                 grating_type=0,
                 pol=0,
                 fourier_order=40,
                 ucell_materials=None,
                 algo='TMM',
                 perturbation=1E-10,
                 device='cpu',
                 type_complex=jnp.complex128,
                 fft_type=0,
                 improve_dft=True,
                 ):

        super().__init__(grating_type=grating_type, n_I=n_I, n_II=n_II, theta=theta, phi=phi, psi=psi, pol=pol,
                         fourier_order=fourier_order, period=period, wavelength=wavelength,
                         ucell=ucell, ucell_materials=ucell_materials,
                         thickness=thickness, algo=algo, perturbation=perturbation,
                         device=device, type_complex=type_complex,)

        self.mode = mode
        self.device = device
        self.type_complex = type_complex
        self.fft_type = fft_type
        self.improve_dft = improve_dft

        self.mat_table = read_material_table(type_complex=self.type_complex)
        self.layer_info_list = []

    def _tree_flatten(self):
        children = (self.n_I, self.n_II, self.theta, self.phi, self.psi,
                    self.period, self.wavelength, self.ucell, self.thickness)
        aux_data = {
            'mode': self.mode,
            'grating_type': self.grating_type,
            'pol': self.pol,
            'fourier_order': self.fourier_order,
            'ucell_materials': self.ucell_materials,
            'algo': self.algo,
            'perturbation': self.perturbation,
            'device': self.device,
            'type_complex': self.type_complex,
            'fft_type': self.fft_type,
        }

        return children, aux_data

    @classmethod
    def _tree_unflatten(cls, aux_data, children):
        return cls(*children, **aux_data)

    @jax.jit
    def solve(self, wavelength, e_conv_all, o_e_conv_all):
        self.kx_vector = self.get_kx_vector(wavelength)

        if self.grating_type == 0:
            de_ri, de_ti, layer_info_list, T1 = self.solve_1d(wavelength, e_conv_all, o_e_conv_all)
        elif self.grating_type == 1:
            de_ri, de_ti, layer_info_list, T1 = self.solve_1d_conical(wavelength, e_conv_all, o_e_conv_all)
        elif self.grating_type == 2:
            de_ri, de_ti, layer_info_list, T1 = self.solve_2d(wavelength, e_conv_all, o_e_conv_all)
        else:
            raise ValueError

        return de_ri.real, de_ti.real, layer_info_list, T1, self.kx_vector

    @jax.jit
    def conv_solve_old(self, ucell):  # TODO: refactor - rename and apply to other parts
        E_conv_all = to_conv_mat_discrete(ucell, self.fourier_order, type_complex=self.type_complex, improve_dft=self.improve_dft)
        o_E_conv_all = to_conv_mat_discrete(1 / ucell, self.fourier_order, type_complex=self.type_complex, improve_dft=self.improve_dft)
        de_ri, de_ti, layer_info_list, T1, kx_vector = self.solve(self.wavelength, E_conv_all, o_E_conv_all)
        return de_ri, de_ti

    @jax.jit
    def conv_solve(self, *args, **kwargs):  # TODO: other backends
        [setattr(self, k, v) for k, v in kwargs.items()]

        if self.fft_type == 0:
            E_conv_all = to_conv_mat_discrete(self.ucell, self.fourier_order, type_complex=self.type_complex, improve_dft=self.improve_dft)
            o_E_conv_all = to_conv_mat_discrete(1 / self.ucell, self.fourier_order, type_complex=self.type_complex, improve_dft=self.improve_dft)
        elif self.fft_type == 1:
            E_conv_all = to_conv_mat_continuous(self.ucell, self.fourier_order, type_complex=self.type_complex)
            o_E_conv_all = to_conv_mat_continuous(1 / self.ucell, self.fourier_order, type_complex=self.type_complex)
        else:
            raise ValueError

        de_ri, de_ti, layer_info_list, T1, kx_vector = self.solve(self.wavelength, E_conv_all, o_E_conv_all)

        self.layer_info_list = layer_info_list
        self.T1 = T1
        self.kx_vector = kx_vector

        return de_ri, de_ti



    @jax.jit
    def conv_solve_spectrum(self, ucell):
        E_conv_all = to_conv_mat_discrete(ucell, self.fourier_order, type_complex=self.type_complex, improve_dft=self.improve_dft)
        o_E_conv_all = to_conv_mat_discrete(1 / ucell, self.fourier_order, type_complex=self.type_complex, improve_dft=self.improve_dft)
        de_ri, de_ti, layer_info_list, T1, kx_vector = self.solve(self.wavelength, E_conv_all, o_E_conv_all)
        return de_ri, de_ti

    def run_ucell(self):
        ucell = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength,
                                          type_complex=self.type_complex)
        if self.fft_type == 0:
            E_conv_all = to_conv_mat_discrete(ucell, self.fourier_order, type_complex=self.type_complex, improve_dft=self.improve_dft)
            o_E_conv_all = to_conv_mat_discrete(1 / ucell, self.fourier_order, type_complex=self.type_complex, improve_dft=self.improve_dft)
        elif self.fft_type == 1:
            E_conv_all = to_conv_mat_continuous(ucell, self.fourier_order, type_complex=self.type_complex)
            o_E_conv_all = to_conv_mat_continuous(1 / ucell, self.fourier_order, type_complex=self.type_complex)
        else:
            raise ValueError

        de_ri, de_ti, layer_info_list, T1, kx_vector = self.solve(self.wavelength, E_conv_all, o_E_conv_all)

        self.layer_info_list = layer_info_list
        self.T1 = T1
        self.kx_vector = kx_vector

        return de_ri, de_ti

    def run_ucell_vmap(self, ucell_list):
        """
        under development
        """

        de_ri, de_ti = jax.vmap(self.conv_solve)(ucell_list)

        return de_ri, de_ti

    def run_ucell_pmap(self, ucell_list):
        """
        under development
        """

        de_ri, de_ti = jax.pmap(self.conv_solve)(ucell_list)
        # de_ri.block_until_ready()
        # de_ti.block_until_ready()

        de_ri = np.array(de_ri)
        de_ti = np.array(de_ti)
        return de_ri, de_ti

    def calculate_field(self, resolution=None, plot=True):

        if self.grating_type == 0:
            resolution = [100, 1, 100] if not resolution else resolution
            field_cell = field_dist_1d(self.wavelength, self.kx_vector, self.n_I, self.theta, self.fourier_order, self.T1,
                                       self.layer_info_list, self.period, self.pol, resolution=resolution,
                                       type_complex=self.type_complex)
        elif self.grating_type == 1:
            resolution = [100, 1, 100] if not resolution else resolution
            field_cell = field_dist_1d_conical(self.wavelength, self.kx_vector, self.n_I, self.theta, self.phi, self.fourier_order, self.T1,
                                               self.layer_info_list, self.period, resolution=resolution,
                                               type_complex=self.type_complex)

        else:
            resolution = [10, 10, 10] if not resolution else resolution
            field_cell = field_dist_2d(self.wavelength, self.kx_vector, self.n_I, self.theta, self.phi, self.fourier_order, self.T1,
                                       self.layer_info_list, self.period, resolution=resolution,
                                       type_complex=self.type_complex)
        if plot:
            field_plot(field_cell, self.pol)
        return field_cell

    # def generate_spectrum(self, wl_list):
    #     ucell = deepcopy(self.ucell)
    #     spectrum_ri_pmap = np.zeros(wl_list.shape)
    #     spectrum_ti_pmap = np.zeros(wl_list.shape)
    #     for i, wavelength in enumerate(range(counter)):
    #         b = i * num_device
    #         de_ri_pmap, de_ti_pmap = jax.pmap(loop_wavelength)(wavelength_list[b:b + num_device],
    #                                                            mat_pmtvy_interp[b:b + num_device])
    #
    #         spectrum_ri_pmap[b:b + num_device] = de_ri_pmap.sum(axis=(1, 2))
    #         spectrum_ti_pmap[b:b + num_device] = de_ti_pmap.sum(axis=(1, 2))
    #
    #     return spectrum_ri_pmap, spectrum_ti_pmap


if __name__ == '__main__':
    pass



from jax import tree_util

from .optimizer import OptimizerJax

tree_util.register_pytree_node(OptimizerJax,
                               OptimizerJax._tree_flatten,
                               OptimizerJax._tree_unflatten)



import jax
import optax

from functools import partial

from ..emsolver.rcwa import RCWAJax


class Grad:

    @staticmethod
    def forward(params, forward, loss):
        result = forward(**params)
        loss_value = loss(result)
        return loss_value

    def _grad(self, params, forward, loss_fn):
        loss_value, grads = jax.value_and_grad(self.forward)(params, forward, loss_fn)
        return loss_value, grads

    def grad(self, pois, forward, loss_fn):
        params = {poi: (getattr(self, poi)) for poi in pois}
        loss_value, grads = jax.value_and_grad(self.forward)(params, forward, loss_fn)
        return loss_value, grads


class OptimizerJax(RCWAJax, Grad):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

    def _tree_flatten(self):  # TODO: check args and kwargs
        children = (self.n_I, self.n_II, self.theta, self.phi, self.psi,
                    self.period, self.wavelength, self.ucell, self.thickness)
        aux_data = {
            'mode': self.mode,
            'grating_type': self.grating_type,
            'pol': self.pol,
            'fourier_order': self.fourier_order,
            'ucell_materials': self.ucell_materials,
            'algo': self.algo,
            'perturbation': self.perturbation,
            'device': self.device,
            'type_complex': self.type_complex,
            'fft_type': self.fft_type,
        }

        return children, aux_data

    @classmethod
    def _tree_unflatten(cls, aux_data, children):
        return cls(*children, **aux_data)

    def gradient_numerical(self):
        pass

    @partial(jax.jit, static_argnums=(3, 4, 5))
    def step(self, params, opt_state, optimizer,  forward, loss_fn):

        loss_value, grads = self._grad(params, forward, loss_fn)

        updates, opt_state = optimizer.update(grads, opt_state, params)
        params = optax.apply_updates(params, updates)
        return params, opt_state, loss_value

    def fit(self, pois, forward, loss_fn, optimizer, iteration=100):
        params = {poi: (getattr(self, poi)) for poi in pois}

        opt_state = optimizer.init(params)

        for i in range(iteration):
            params, opt_state, loss_value = self.step(params, opt_state, optimizer, forward, loss_fn)
            if i % 1 == 0:
                print(f'step {i}, loss: {loss_value}')
